#include "chpoxSupport.h"

/**
 * Default constructor for chpoxSupport class 
 */
chpoxSupport::chpoxSupport()
{
	m_isActive = isEnabled();
}

/**
 * checks if chpox is installed and enabled on the local node
 * @return true if chpox is installed and enabled on local node; false otherwise
 */
bool chpoxSupport::isEnabled()
{
	if(m_util.readLineFromFile(CHPOX_INFO_FILE) == UTILS_IO_ERROR)
	{
		return false;
	}
	return true;
}

/**
 * invokes GNU's ldd program to retrieve the list of required libs for a given program.
 * @param filePath is the path to the input process
 * @return a list with required libs for the input program; if an error occurred the returned list is empy (size==0)
 */
list<string> chpoxSupport::getLibsList(string filePath)
{
	//parse the command ("ldd PROGRAM" is tokenized and saved in a linked list)
	list<string> cmdsList = m_util.parseCommand("ldd "+filePath);

	 //execute the command (in a separate child process) and get+return the output of the
	 //executed command (in this case we don't require real-time output => second param is set to NULL)
	return m_util.execCommandAndReturnChildOutput(cmdsList, NULL);
}

/**
 * searches the list of previously saved (==yet registered) procs for a specific program.
 * If the answer is not succesfull, creates a new entry for the proc; otherwise
 * update the correct entry.
 * @param pid is the PID  of the program to be searched/updated in the saved proc list
 * @param exe is the name of the program to be searched/updated in the saved proc list
 * @param requiredLibs is a list of libs required by the process. It's used only on inserting phase.
 * @param dumpFile is the name for the new dump file.
 */
void chpoxSupport::updateList(pid_t pid, string exe, list<string> requiredLibs, string dumpFile)
{
	list<chpoxProc>::iterator iter;

	//search the input process in the list of saved (==yet registered) procs
	for(iter=m_registeredProcs.begin(); iter!=m_registeredProcs.end(); ++iter)
	{
		if(iter->m_pid==pid && iter->m_exe==exe) //if the given proc is found (==the related entry was found)...
		{
			cout << "[CHP]: proc yet registered: proc entry updated." << endl;
			iter->setNewDump(dumpFile); //set a new dump filename for the found entry
			cout << "[CHP]: number of saved dumps for the registered proc: " <<  iter->m_dumpsNumber << endl;
			return; //exit function
		}
	}

	//if we are here, the input process wasn't found => we must create a new entry for the input process
	//in the list of saved (==yet registred) procs.
	chpoxProc chp(pid, exe); //create the new entry
	chp.setNewDump(dumpFile); //set a new dump filename for the new created entry
	chp.setRequiredLibs(requiredLibs); //set the libs required by the program
	cout << "[CHP]: added a new entry to the list of registered procs." << endl;
	cout << "[CHP]: number of saved dumps for the registered proc: " <<  chp.m_dumpsNumber << endl;
   	m_registeredProcs.push_back(chp); //add the new entry to the list of saved (==yet registered) procs
	cout << "[CHP]: the list of registered procs contains " << m_registeredProcs.size() << " entries." <<  endl;
}

/**
 * forces the dump (==checkpoint) for the given process
 * @param pid is the PID of the process to be dumped (checkpointed)
 */
void chpoxSupport::dumpRegisteredProcess(pid_t pid)
{
	errno=0;
	if(kill(pid, CHPOX_DUMP_SIGNAL)==0)  //if we succeeded sending the "force-proc-dump" SIGNAL to the given process...
	{
		cout << "[CHP]: dump signal (" << CHPOX_DUMP_SIGNAL<< ") succesfully sent to target process (pid=" << pid << ")" << endl;
	}
	else //if the SIGNAL didn't reach the given process...
	{
		cout << "[CHP]: error:  dump signal (" << CHPOX_DUMP_SIGNAL<< ") NOT succesfully sent to target process (pid=" << pid << ")" << endl;
		cout << "[CHP]: error description: " << errno << endl;
	}
}

/**
 * used to register and dump the given process
 * @param pid is the PID for the process to be registered
 * @return true if all went right; false otherwise
 */
bool chpoxSupport::registerAndDumpProcess(pid_t pid)
{
	string cmdline;
	string libPath;;
	string chpoxLine;
	string dumpFilePath;
	string exe;
	list<string> libsPaths;
	list<string> cleanedLibsPaths;
	list<string> cmdsList;
	list<string>::iterator iter;
	map<string, string>::iterator mapIter;

	if(m_isActive==false)
	{
		cout << "[CHP]: error: chpox support is disabled!" << endl;
		return false;
	}

	try
	{
	        /**
		 * STEP #1: read exe file in /proc/[pid] directory
		 */
		exe = m_util.getFilenamePointedByLink("/proc/"+m_util.longToString(pid)+"/exe");
		if(exe==UTILS_IO_ERROR)
		{
			cout << "[CHP]: error: cannot find process with PID="<< pid << endl;
			return false;
		}
		m_util.trim(exe);
		cout << "[CHP]: pid: [" << pid << "] cmdline: ["<< exe << "]" << endl;

		 /**
		 * STEP #2: get and register all the libs required by the given program
		 */
		libsPaths = getLibsList(exe); //get the list of the libs required by the input process
		cout << "[CHP]: registering the libs needed by the program "+exe+"." << endl;
		for(iter=libsPaths.begin(); iter!=libsPaths.end(); ++iter)
		{
			//this stuff cleans the output of GNU ldd programs so we can
			//get only the cleaned lib paths.
			libPath=*iter;
			libPath= libPath.substr(0,libPath.find("0x")-1);
			libPath= libPath.substr(libPath.find("=>")+3);
			m_util.trim(libPath);

			//add the cleaned path to a list with all the cleaned libs paths
			cleanedLibsPaths.push_back(libPath);
			cout << "       => " << libPath  << endl;

			//open the chpox file used to save the lib paths (so they will be registered)
			m_util.openFile("/proc/chpox/libs", true);
			if(m_util.writeToOpenFile("+"+libPath)==false)
			{
				cout << "[CHP]: error: error writing on file /proc/chpox/libs" << endl;
				return false;
			}
			m_util.closeFile(); //close the chpox file
		}
		cout << "[CHP]: the program needs " << cleanedLibsPaths.size() << " libs." << endl;

		/**
		 * STEP #3: register the process
		 */
		string currTime = m_util.getLocalTime();

		string::size_type f= exe.rfind("/");
		cmdline = exe.substr(f+1);
		cout << "[CHP]: registering the app...." << endl;

		//create the name (and the path) for the new dump file
		//FORMAT:  PATH/EXENAME_YYYYMMDD_HHMMSS_PIDT.dump
		dumpFilePath   = CHPOX_DUMP_DIR;
		dumpFilePath += "/";
		dumpFilePath += cmdline;
		dumpFilePath += currTime;
		dumpFilePath += "_";
		dumpFilePath += m_util.longToString(pid);
		dumpFilePath += ".dump";


		//create the string for registering the process (as required by chpox): 4 params required
		chpoxLine   = m_util.longToString(pid);                                  //#1 PID
		chpoxLine += ":";
		chpoxLine += m_util.intToString(CHPOX_DUMP_SIGNAL); //#2 SIGNAL TO STOP
		chpoxLine += ":";
		chpoxLine += CHPOX_REGISTER_MECHANISM;              //#3 WHAT TO SAVE (9=register children procs also)
		chpoxLine += ":";
		chpoxLine += dumpFilePath;                                                //#4 WHERE TO SAVE (DUMP FILEPATH)
		cout << "       => " << chpoxLine  << endl;

		m_util.openFile("/proc/chpox/register", true);
		if(m_util.writeToOpenFile(chpoxLine)==false) //write
		{
			cout << "[CHP]: error: error writing on file /proc/chpox/register" << endl;
			return false;
		}
		m_util.closeFile();

		/**
		 * STEP #4: checkpoint the process
		 */
		dumpRegisteredProcess(pid);

		/**
		 * STEP #5: check if the dump file was correctly generated (==dump file exists) and
		 *                 if the dump file is not CORRUPTED
		 */
		cout << "[CHP]: wait " << CHPOX_REFRESH_TIME << " ms before continuing..." << endl;
		for(int i=CHPOX_REFRESH_TIME;i>0;i--)
		{
			printf("[CHP]: countdown: %d \r",i);
			fflush(stdout);
			sleep(1);
		}

		errno=0;
		struct stat fileToBeSearched;
		if(stat(dumpFilePath.c_str(), &fileToBeSearched)==-1 && errno== ENOENT)
		{
			cout << "[CHP]: error: dump file was NOT created! (file=" << dumpFilePath << ")" << endl;
			return  false;
		}

		string integrityStatus = m_util.readLineFromFile("/proc/chpox/info");
		string::size_type pos1 = integrityStatus.find("[");
		string::size_type pos2 = integrityStatus.find("|");
		string::size_type pos3 = integrityStatus.find("]");
		if(integrityStatus.substr(pos1+1,1)=="E")
		{
			cout << RED << "[CHP]: integrity status: CORRUPTED!" << NORMAL << endl;
			return false;
		}
		else
		{
			cout << GREEN << "[CHP]: integrity status: OK! => " << integrityStatus.substr(pos2+1,pos3-pos2-1) << " child process(es)" << NORMAL << endl;
		}

		//If we're here, all went right!

		/**
		 *  STEP #6: if  the process was yet registered (an entry is found in the list of saved procs), update
		 *                 its entry; otherwise create a chpoxProc object (to remember registered process and libs for easy removal)
		 *                 then add  to a global list of saved procs (chpoxProc objects)
		 */
		cout << "[CHP]: dump file correctly created (file=" << dumpFilePath << ")" << endl;
		updateList(pid, exe, cleanedLibsPaths, dumpFilePath);

		/**
		 *  STEP #7: check if the program has a special associated dump behaviour
		 *                 (this is true if the program appears in the m_chpSupportConfiguration map)
		 */
		mapIter = m_chpSupportConfiguration.find(cmdline);
		if(mapIter!=m_chpSupportConfiguration.end()) //if the process is listed in the special chpoxsupport file...
		{
			if(mapIter->first==m_util.intToString(CHPOX_DUMP_PERSISTENT)) //if the process needs persistent (==periodic) dump...
			{
				cout << RED << "[CHP]: program " << cmdline << " requires persistent dump." << NORMAL << endl;
				appendToPendingDumpsQueue(pid); //...appending again the process PID to the end of the pending-dumps queue
			}
		}
	}
	catch(...) //if any error/exception occurred...
	{
		return false;
	}
	return true;
}

/**
 * prints the contents of the list of saved (==yet registered) procs
 */
void chpoxSupport::printRegisteredProcList()
{
	int i = 1;
	int counter = 0;
	list<chpoxProc>::iterator iter;
	list<string>::iterator iter2;

	//for each entry in the list of saved (==yet registered) procs
	for(iter=m_registeredProcs.begin(); iter!=m_registeredProcs.end(); ++iter)
	{
		++counter;
		cout << "       => Entry #" << counter  << endl;

		i=1;
		cout << "          Required libs ("<< iter->m_requiredLibs.size() <<" libs found):"  << endl;
		for(iter2=(iter->m_requiredLibs).begin(); iter2!=(iter->m_requiredLibs).end(); ++iter2)
		{
			cout << "          - (#" << i << ") " << *iter2  << endl;
			++i;
		}

		i=1;
		cout << "          Dump filenames ("<< iter->m_dumpsNames.size() <<" dump files found):"  << endl;
		for(iter2=(iter->m_dumpsNames).begin(); iter2!=(iter->m_dumpsNames).end(); ++iter2)
		{
			cout << "          - (#" << i << ") " << *iter2  << endl;
			++i;
		}
		cout << "" << endl;
	}
}

/**
 * used to start the internal thread
 */
void chpoxSupport::run()
{
	pid_t p;
	list<pid_t>::iterator iter;

	//repeat forever...
	while(true)
	{
		//check if the queue contains at least one pending dump-request...
		if(m_pendingDumpsQueue.size()>0)
		{
			p = m_pendingDumpsQueue.front(); //get the pid of the first element
			m_pendingDumpsQueue.pop_front(); //remove the first element (as we're processing it)

			cout << "[CHP]: try registering process with pid=" << p << "            " << endl;
			cout << "[CHP]: queue contains " << m_pendingDumpsQueue.size()+1 << " pids." << endl;

			//try to register and dump the (removed) first element of the queue
			if(registerAndDumpProcess(p)==false) //if any error war raised...
			{
				cout << GREEN << "[CHP]: error: error while registering/checkpointing the process with pid="<< p << NORMAL << endl;
			}
			else //if all went right...
			{
				cout << RED << "[CHP]: process successfully registered and dumped  (pid="<< p << ")" << NORMAL << endl;
			}
		}
		//wait 5 secs between a queue-check/dump-request-execution and another one
		//(so this thread:
		//1) uses a small amount of resources (no busy-waiting)
		//2) doesn't interfere with the main process of the daemon
		//3) allows to append as many dump-requests as needed
		//    (so guarantees a quick'n safe way to process dump-requests.)
		Thread::sleep(5000);
	}
}

/**
 * used to append a new dump request to the specific dump-requests queue
 * @param newPid is the pid of the process to be added to the queue of dump-requests
 * @return true if all goes right; otherwise return false
 */
bool chpoxSupport::appendToPendingDumpsQueue(pid_t newPid)
{
	if(m_isActive==false)
	{
		cout << "[CHP]: error: chpox support is disabled!" << endl;
		return false;
	}

	cout << "[CHP]: appending new process to the queue  (pid="<< newPid << ")" << endl;

	//append the pid of the process to be added to the queue of dump-requests
	m_pendingDumpsQueue.push_back(newPid);
	return true;
}

/**
 * search for migrates processes on local node.
 * Migrated procs will be added to a special list
 */
void chpoxSupport::getListOfMigratedProcs()
{
	long nodeId = 0;
	long foundMigratedProc;
	string ret;
        string tmpstr;
	string currFile;

	try
	{
		DIR *directory; //directory handle
		struct dirent *info; //directory entry
		directory = opendir("/proc"); //open /proc directory
		if(directory==NULL) //if we cannot access the /proc directory...
		{
			cout << "[CHP]: error: cannot access /proc directory!" << endl;
			return;
		}

		//clear the special list
		m_listOfMigratedProcs.clear();

		//for each proc directory in /proc directory...
		while((info=readdir(directory))!=NULL)
		{
			currFile = string(info->d_name); //get a file name
			tmpstr = "/proc/"+currFile+"/where"; //create the complete path to the file

			printf("[CHP: inspecting dir %s...               \r", tmpstr.c_str()); //show the directory we're inpecting...
			fflush(stdout);
			sleep(1);	//wait one ms => so you can see what's doing... ;)

			//try to read the special "where" files in the proc directory (openMosix special file)
			nodeId = m_util.stringToLong(m_util.readLineFromFile(tmpstr).c_str());

			//if the "where" special file exists and the special file reveal the proc was migrated...
			//(a proc is migrated if contains a number != 0  => 0 means the proc is local!)
			if(nodeId>0)
			{
				//if we're here, a migrated proc is detected!
				foundMigratedProc = m_util.stringToLong(currFile.c_str());
				printf("[CHP]: found a migrated process! (pid=%ld)\n", foundMigratedProc);
				fflush(stdout);

				//append the found proc the list of migrated procs
				m_listOfMigratedProcs.push_back(foundMigratedProc);
			}
		}

		//we've finished!
		printf("[CHP: job done!                                                               \n");

	}
	catch(...)
	{
		cout << "[CHP]: error: exception raised while creating the list of migrated procs!" << endl;
		return;
	}
}

/**
 * reads the file with the procs to be dumped
 * @param chpFile is the path to the chpox-support configuration file 
 */
void chpoxSupport::fillChpSupportConfiguration(string chpFile)
{
	cout << "[CHP]: loading CHP support configuration from file..." << endl;
	cout << "[CHP]: opening file: \"" << chpFile << "\""  << endl;
	multimap<string, string> fromFile = m_util.parseSpecialFile(chpFile,":");
	map<string, string>::iterator iter;

	//multimap to map conversion
	for(iter=fromFile.begin(); iter!=fromFile.end(); ++iter)
	{
		m_chpSupportConfiguration.insert(make_pair(iter->first, iter->second));
	}
	cout << "[CHP]: CHP support configuration map succesfully filled." << endl;
}

/**
 * prints the contents of the procs-to-be-dumped queue
 */
void chpoxSupport::printChpSupportConfiguration()
{
	//if the map is empty, return immediately...
	if(m_chpSupportConfiguration.size()==0)
	{
		return;
	}

	cout << "[CHP]: contents of CHP-support configuration map..." << endl;
	map<string, string>::iterator iter;

	for(iter = m_chpSupportConfiguration.begin(); iter != m_chpSupportConfiguration.end(); ++iter)
	{
		cout << "       -> " << iter->first << ":"<< iter->second << endl;
	}
}

/**
 * convert a string to its autodump value
 * @param toBeConverted is the string to be converted
 * @return the associated autodump type (if available,otherwise returns CHPOX_DUMP_ONCE value)
 */
chpox_autodump_t chpoxSupport::stringToAutodumpType(string toBeConverted)
{
        if (toBeConverted=="CHPOX_DUMP_PERSISTENT")
        {
                return CHPOX_DUMP_PERSISTENT;
        }
        return CHPOX_DUMP_ONCE; //default value is CHPOX_DUMP_ONCE
}

/**
 * used to restore a process via its dump file
 * @param pidOfProcToBeRestored is the pid of the porcess to be restored
 * @param numOfDumpFileToBeUsed is the path to the dump file to be used for restoring
 * @param killProcessIfStillRunning is used to force the kill of the process, if it's still running
 * @return a code if any error occurred; otherwise return no message
 */
string chpoxSupport::restoreProcess(pid_t pidOfProcToBeRestored, unsigned int numOfDumpFileToBeUsed, bool killProcessIfStillRunning)
{
	unsigned int i = 0;
	chpoxProc* chp = NULL;
	string foundDumpfile = "<ERROR>";
	list<chpoxProc>::iterator iter;
	list<string>::iterator iter2;

	if(killProcessIfStillRunning==true)
	{
		kill(pidOfProcToBeRestored, SIGKILL);
		cout << "[CHP]: dump signal (SIGKILL) successfully sent to target process (pid=" << pidOfProcToBeRestored << ")" << endl;
	}

	//for each entry in the list of saved (==yet registered) procs
	for(iter=m_registeredProcs.begin(); iter!=m_registeredProcs.end(); ++iter)
	{
		//cout << "#"<< iter->m_pid << "# <-> #" << pidOfProcToBeRestored << "#" << endl;
		if(iter->m_pid==pidOfProcToBeRestored)
		{
			chp = new chpoxProc(*iter);
			break;
		}
	}

	if(chp==NULL || numOfDumpFileToBeUsed>chp->m_dumpsNames.size() || numOfDumpFileToBeUsed<0)
	{
		return "-40";
	}

	for(iter2=(chp->m_dumpsNames).begin(); iter2!=(chp->m_dumpsNames).end(); ++iter2)
	{
		++i;
		if(i==numOfDumpFileToBeUsed)
		{
			foundDumpfile = *iter2;
			break;
		}
	}

	if(foundDumpfile=="<ERROR>")
	{
		return "-41";
	}

	cout << "[CHP]: restoring the process..." << endl;
	cout << "[CHP]: invoking \"ld-chpox -v " << foundDumpfile << "\"" << endl;
	list<string> cmdsList = m_util.parseCommand("ld-chpox -v "+foundDumpfile);
	m_util.execCommandAndReturnChildOutput(cmdsList, NULL);  //execute the command (second param is set to NULL => no real-time output)
	for(iter2=cmdsList.begin(); iter2!=cmdsList.end(); ++iter2)
	{
		cout << "       -> "  << *iter2 << endl;
	}

	cout << "[CHP]: job done! " << endl;
	return "";

}

/**
 * used to get the list of available dump-files
 * @return a string-based list of available dump-files
 */
string chpoxSupport::getListOfAvailableDumps()
{
	string app;
	string date;
	string time;
	string pid;

	string ret;
        string tmpstr;
	string fileName;

	DIR *directory; //directory handle
	struct dirent *info; //directory entry
	int n=0;
	cout << "[CHP]: contents of \"" << CHPOX_DUMP_DIR << "\" directory" << endl;
	try
	{
		directory = opendir(CHPOX_DUMP_DIR); //open HPC iface nodes directory
		if(directory==NULL)
		{
			//cannot access CHPOX_DUMP_DIR directory
			ret =   "<ERROR>: can't open ";
			ret += CHPOX_DUMP_DIR;
			ret += " directory.";
			return ret;
		}

		while(((info=readdir(directory))!=NULL)) //read the directory (read each file included in that directory)
		{
			fileName = string(info->d_name); //get a file name
			if(fileName.size()>8 && fileName.substr(fileName.size()-5)==".dump")
			{
				++n;
				/*StringTokenizer st(fileName,"_");
				StringTokenizer::iterator i;
				app = *i;
				date = *(++i);
				time = *(++i);
				pid = *(++i);*/
				m_foundDumps.insert(make_pair(fileName,""));
			}
		}

		closedir(directory); //close HPC iface nodes directory

		ret = "OK!";
	}
	catch(...)
	{
		ret="<ERROR>: exception occurred.";
	}
	return ret;
}

