/***************************************************************************
*                          connHandler.cpp  -  description
*                             -------------------
*    begin                : Sam Feb  1 17:08:52 CET 2003
*    copyright            : (C) 2003 by the connHandler group at savannah.org/
*    email                : rpons@rinu.org, brian@axehind.com, jp80@libero.it, mosixview@t-online.de
****************************************************************************/

/**
 * NOTES:
 * Created by openMosixview
 * Last update by rejected
 *
 * Contains the code to handle a single client connection
 */

#include "connHandler.h"

/**
 * constructor for connHandler class
 * @param server is a reference to the server socket
 */
connHandler::connHandler(ncSocket& server) : TCPStream(server), Thread()
{
	//cout << "<CONN HANDLER> Creating a new session for a client." << endl;
	setCompletion(true); //set non-blocking-socket mode
	m_remoteClientIP.append(server.m_lastAcceptedClientIP);
	m_requireCleanup=false;
}

/**
 * init function for connHandler class
 * @param gi is a reference to global (singleton) globalInitializer instance
 * @param slog is a pointer to the global syslogger object
 * @param socketTimeout is the socket timeout for this object
 * @param localIP contains the local IP (C string)
 * @param localHostname is the name of the local machine (C string)
 */
void connHandler::init(globalInitializer* gi, sysLog &slog, int socketTimeout, const char *localIP, const char *localHostname)
{
	m_localNodeIP.append(localIP); //set local node IP address
	m_localNodeHostname.append(localHostname); //set local node hostname address
	m_lastRequest.erase(); //clean up
	m_socketTimeout = socketTimeout; //store socket timeout value
	setTimeout(m_socketTimeout); //set socket timeout for this object

	m_slog         = new sysLog(slog);
	m_globalInit   = gi; //store a reference to the global csColector object

	m_localNodeID = m_globalInit->m_localNodeID; //set local node ID
	if(m_localNodeID==UTILS_IO_ERROR)
	{
		m_localNodeID="-1";
	}
}

/**
 * destructor for class connHandler
 */
connHandler::~connHandler()
{
	//delete m_slog; //used to write datas to system logger
	//delete m_globalInit;
	cout << "[INF]: cleaning up after closing connection to " << m_remoteClientIP << "." <<  endl;
}

// methods implementations
// connHandler methods

/**
 * thread code. Main method for class connHandler.
 * all connHandler-to-client communication code is here!
 */
void connHandler::run()
{
	bool doContinue = true;		//loop (while) control variable
	//send program banner to the client
	sendBannerToClient();
	setError(true);

	try
	{
		//repeat 'till the exit is forced
		while (doContinue==true)
		{
			//max request size:

			if(m_lastRequest=="") //if this var doesn't contains anything -> wait for a client request
			{
				//request a command from the client
				*tcp() << "gomd@" << m_localNodeHostname << " # ";
				sync();

				//read the request from the client
				this->readLine(m_buffer, CONN_HANDLER_MAX_REQUEST_LENGTH, m_socketTimeout);
				if(strlen(m_buffer)==0)
				{
					cout << "[ERR] buffer is empty!" << endl;
				}
				else
				{
					m_requestHistory.append(m_buffer); //append the request to the request history
				}
			}
			else //if this var contains something -> user wants to repeat last request
			{
				int len = strlen(m_lastRequest.c_str());
				strncpy(m_buffer,m_lastRequest.c_str(),len);
			}

			//clean up member vars before receiveing a new request
			m_rCmd.erase();
			m_rRequest.erase();
			m_rParam.erase();
			m_rFunctionByNumber = ANALYZER_UNVALID_OPTION;


			//1)parse the request
			parseRequest(m_buffer, m_rCmd, m_rRequest, m_rParam); //if the request is valid...


			//2)if necessary convert short-format request to long-format one (=fix short-format ones!)
			fixShortFormatRequest(m_rCmd, m_rRequest, m_rParam);


			//3)analyze the request
			m_rFunctionByNumber = analyzeOption(m_buffer, m_rCmd, m_rRequest, m_rParam);
			if(m_rFunctionByNumber==ANALYZER_UNVALID_OPTION)
			{
				*tcp() << "<ERR>: request is invalid!" << endl;
			}
			else
			{
				m_answerBuffer.erase(); //clean up the answer buffer

				//4)execute the request (m_answerBuffer will be filled accordingly)
				doContinue = splitComplexRequest(m_rFunctionByNumber, m_rParam);


				//5)If the answer buffer is not empty, send it to the client
				//NOTE: if the buffer is empty => something cleaned up it and sent
				//its answer directly to client (i.e. "get help" request)
				if(m_answerBuffer.size()!=0)
				{
     					*tcp() << m_answerBuffer << endl; //send the answer to the client
				}
				else
				{
					*tcp() << "[ERR]: openMosix not started or node not configured." << endl;
				}
			}
			this->yield(); //give priority to another thread
		}
	}
	catch(Socket *socket) //catch conn errors/exceptions
        {
		int err = socket->getErrorNumber(); //get the id associated to the socket error/exception
		switch (err)
		{
			case Socket::errConnectRefused:
			{
				m_slog->log(LOG_INFO,"[ERR]: connection refused!");
				break;
			}
			case Socket::errTimeout:
			{
				m_slog->log(LOG_INFO,"[ERR]: timeout expired!");
				break;
			}
			case Socket::errSuccess:
			{
				m_slog->log(LOG_INFO,"[ERR]: connection closed by remote host.");
				break;
			}
			case Socket::errInput:
			{
				m_slog->log(LOG_INFO,"[ERR]: connection closed by remote host.");
				m_slog->log(LOG_INFO,"[ERR]: (remote request executed.)");
				break;
			}
			default:
			{
				m_slog->log(LOG_INFO,"[ERR]: connection error occurred!");
				cout << "[ERR]: error code: " << err << endl;
				break;
			}
		}
	}
	catch(...) //catch every other excption/error
	{
		m_slog->log(LOG_INFO,"[ERR]: unknown exception was raised!");
		*tcp() << "[ERR]: critical error occurred!" << endl;
	}

	cout << "[INF]: connection to " << m_remoteClientIP << " closed." <<  endl;

	m_requireCleanup = true; //the chSimpleGC will be able to cleanup this object
	TCPStream::endStream();  //close the conn (close the stream)
	Thread::setCancel(Thread::cancelDefault);
	Thread::exit();          //kill the thread

	if(this!=NULL)
	{
		cout << GREEN << "I'm living!!!" << NORMAL << endl;
	}
}

/**
 * used to get a boolean value used by chSimpleGC to decide if this object can be safely cleaned up
 * @return a boolean value used by chSimpleGC to decide if this object can be safely cleaned up
 */
bool connHandler::isCleanupRequired()
{
	return m_requireCleanup;
}


/**
 * used to parse a request (splits the request in a command, in a request, in the target-nodes)
 * @param clientRequest is the request from the client
 * @param rCmd is a string where the command will be placed
 * @param rRequest is a string where the request will be placed
 * @param rParam is a string where the target-nodes will be placed
 */
void connHandler::parseRequest(char* clientRequest, string& rCmd, string& rRequest, string& rParam)
{
	int len = strlen(clientRequest)-1;
	*(clientRequest+len) = ' ';

	StringTokenizer st(clientRequest," ");
	StringTokenizer::iterator i;

	try
	{
		i=st.begin();
		m_rCmd = *i;
		m_rRequest = *(++i);
		m_rParam = *(++i);
	}
	catch(StringTokenizer::NoSuchElementException ex){} //ignore exceptions
}

/**
 * used to convert short-format requests to long-format requests (properly, it fixes them!)
 * @param cmd is the string with the command token (yet parsed in parseRequest function)
 * @param request is the string with the the request token (yet parsed in parseRequest function)
 * @param param is the string with the the param token (yet parsed in parseRequest function)
 */
void connHandler::fixShortFormatRequest(string& cmd, string& request, string& param)
{
	/*
	SPECIAL COMMANDS SUPPORT
	for _some_ commands we provide a short request syntax:
	short format will be automagically translated in long format

	example: the following two requests are equivalent
	1) help 1 (SHORT FORMAT)
	2) get help 1 (LONG FORMAT)

	THE PROBLEM
	if the user enter a special short-format request ,
	the request tokens result "shifted"
	1)_real_ request is placed in var "cmd" (instead of "request" var)
	2)_real_ param is placed in var "request" (instead of "param" var)
	3)obviously "param" var is empty

	THE SOLUTION
	the solution is simple:
	we move request and param to the right variables and
	we add "get" into "cmd" variable.

	  * uncorrect!
	  CMD  REQUEST PARAM
	  help 1       NULL

	  * fixed! (we add the correct command for the request and
	    shift values to the right)
	  CMD  REQUEST PARAM
	  get  help    1

	after this "correction" the request will be normally processed

	*/
	if(cmd=="help" || cmd=="quit")
	{
		param = request; //shift to the right
		request = cmd; //shift to the right
		cmd = "get"; //add the correct command for these requests ("get")
	}
	else if(cmd=="cmd" || cmd=="con" || cmd=="cwc" || cmd=="dump")
	{
		param = request; //shift to the right
		request = cmd; //shift to the right
		cmd = "exec"; //add the correct command for these requests ("get")
	}

}

/**
 * used to analyze and check a request (splitted in a command, in a request, in a param)
 * @param clientRequest is a string with the request received from the client (string is in "pure C" style)
 * @param cmd is a command string to check
 * @param request is a request string to check
 * @param param is a param string to check
 * @return a integer if the request is valid (to be passed to splitComplexRequest() method to call a specific function). If the request is invalid, return ANALYZER_UNVALID_OPTION
 */
int connHandler::analyzeOption(char* clientRequest, string& cmd, string& request, string& param)
{
	string currRule_command;
	string currRule_longOpt;
	string currRule_shortOpt;
	string currRule_mnem;
	string currRule_validvals;

	//Check if a request valid by checking all the rules (if necessary)
	for(unsigned int i=0;i< m_globalInit->m_chSupport->m_validRules.size();i++)
	{

		currRule_command   = m_globalInit->m_chSupport->m_validRules[i].m_command;
		
		//Check if entered Command is valid or marked as "unuseful" (NA)
		if(currRule_command==cmd || currRule_command==NA)
		{
			currRule_longOpt   = m_globalInit->m_chSupport->m_validRules[i].m_longOption;
			currRule_shortOpt  = m_globalInit->m_chSupport->m_validRules[i].m_shortOption;
			currRule_mnem      = m_globalInit->m_chSupport->m_validRules[i].m_mmemonic;
			
			//Check if entered Request is valid
			if(currRule_longOpt==request ||
			   currRule_shortOpt==request ||
			   currRule_mnem==request)
			{
				
				currRule_validvals = m_globalInit->m_chSupport->m_validRules[i].m_validValues;
				
				//Checking request using its valid values
				if(currRule_validvals==NA) //if rule has no specified valid values (NA ="not available")...
				{
					if(m_rParam.size()>0) //if a param was entered (unuseful in this context because it is marked as "NA")...
					{
						m_rParam.erase(); //clean up param (=ignore param)
					}
					return m_util.stringToInt(currRule_mnem.c_str());
				}
				else if(currRule_validvals==NODELIST) //if rule wants a valid nodes list...
				{
					if (param.size()==0) //if nodelist is empty...
					{
						m_rParam=m_localNodeID; //use local node id as target node
					}

					//REMOVE_FOR_DEBUG *tcp() << "<ANALYZER>: VALUE VERIFIED (NODELIST): OK!"<< endl;
					return m_util.stringToInt(currRule_mnem.c_str());
				}
				else if(currRule_validvals==REMOTE_OR_LOCAL_NODE) //if rule wants a valid nodes list...
				{
					if (param.size()==0) //if nodelist is empty...
					{
						m_rParam=m_localNodeID; //use local node id as target node
					}
					else
					{
						m_remoteRequest = true;
					}
					return m_util.stringToInt(currRule_mnem.c_str());
				}
				else if(currRule_validvals==NUMBER) //if rule wants a number..
				{
					if(param.size()==0)
					{
						return ANALYZER_UNVALID_OPTION;
					}

					for(unsigned int j=0;j<param.size();j++) //check if entered param is a valid number (checks char-by-char)
					{
						if(!isdigit(param.at(j))) //if param contains a non-digit char => this is not a number => invalid option!
						{
							return ANALYZER_UNVALID_OPTION;
						}
					}

					//if we're here, all went right
					return m_util.stringToInt(currRule_mnem.c_str());
				}
				else if(currRule_validvals==STRING) //if rule wants a valid nodes list...
				{
					m_rParam=m_buffer; //use local node id as target node
					return m_util.stringToInt(currRule_mnem.c_str());
				}
				else if(currRule_validvals.find(param)!=string::npos) //if param contains a valid option					
				{	//(if we're here, rule wants a pre-defined list of valid values)
					return m_util.stringToInt(currRule_mnem.c_str());
				}
			}
		}
	}
	//if we're here, an error was raised
	return ANALYZER_UNVALID_OPTION;
}

/**
 * used to split a valid request in multiple simple requests (if necessary)
 * @param param is a param string to be splitted (if necessary: i.e. a valid nodes list)
 * @param functionByNumber is the id number of a function to be called (will be passed to executeSingleRequest() method)
 * @return true if no error was raised during request(s) execution, false otherwise
  */
bool connHandler::splitComplexRequest(int functionByNumber,string& param)
{
	bool foundRange=false;
	char c;
	int firstRangeNode=0; //used to store nodes range values
	int lastRangeNode=0;
	string buffer;

	//cout << "SPLITTER functionByID:"<< m_util.intToString(functionByNumber) << " param:" << param << endl;

	//if we have no complex param (i.e.: a node list)...
	if(param.size()==0 || param=="") //if this is a simple command (without node-targets, for example)
	{
		//execute this simple request and exit.
		return executeSingleRequest(functionByNumber, "");
	}
	else if(param.substr(0,8).find("cmd")!=string::npos)
	{
		//cout << "CMD TO BE EXECUTED!" << endl;
		return executeSingleRequest(functionByNumber, param);
	}
	else if(param.substr(0,8).find("con")!=string::npos)
	{
		//cout << "CMD TO BE EXECUTED!" << endl;
		return executeSingleRequest(functionByNumber, param);
	}
	else if(param.substr(0,8).find("cwc")!=string::npos)
	{
		//cout << "CWC TO BE EXECUTED!" << endl;
		return executeSingleRequest(functionByNumber, param);
	}


	//special macro expansion
	//if the user specified ALL as param, this macro will be
	//expanded as the complete oM node list
	//i.e. "ALL" => "1,2,3,4,5,..."
	string tmp = param;
	m_util.ucase(tmp);
	if (tmp=="ALL")
	{
		param = m_globalInit->m_omInfo.getNodesList(false);
	}

	//check if we have a (complex) node list (it contains ',' and/or '-' symbols)
	if(param.find(",")!=string::npos ||
	   param.find("-")!=string::npos)
	{
		//check for required final comma character
		//(we add this comma so param tokenizing will be easy -> it wants a comma
		// at the end of the param string!)
		if (param.at(param.size()-1)!=',')
		{
			param.append(",");
		}
	}
	else //if we are here we have a simple request with a single param (can be also a request with a single node as param)
	{
		//execute this simple request and exit.
		return executeSingleRequest(functionByNumber, param);
	}

	/*
		COMPLEX REQUEST HANDLING (==MORE CMD EXECUTIONS)
		i.e. node lists
	*/

	//if we are here, we got a complex param (i.e.: a node list)...
	//check every character in the param string...
	for(unsigned int i=0;i<param.size();i++) //loop until it reaches the end of string
	{
		c = param.at(i);
		if(c=='-') //if we find the range separator...
		{
			foundRange = true; //we've just found a range of nodes
			firstRangeNode=m_util.stringToInt(buffer); //buffer contains yet the first node of the range!
			buffer.erase(); //cleanup the buffer
		}
		else if(c==',') //if we find the node separator...
		{
			if(foundRange == true) //if a range was already found...
			{
				lastRangeNode = m_util.stringToInt(buffer);  //buffer contains yet the last node of the range
				foundRange=false;

				if(firstRangeNode==-1 || lastRangeNode==-1) //check we've correct range values
				{
					*tcp() << "<ERR>: cannot execute malformed nodes-range request" << endl;
				}
				else
				{
					if (firstRangeNode>lastRangeNode)
					{
						int tmp = firstRangeNode;
						firstRangeNode=lastRangeNode;
						lastRangeNode=tmp;
					}

					for(int z=firstRangeNode;z<=lastRangeNode;z++)
					{
						executeSingleRequest(functionByNumber, m_util.intToString(z));
					}
				}
			}
			else //if we are here, we haven't got a range of nodes
			{
				//execute the request for the found node
				if(!executeSingleRequest(functionByNumber, buffer)) //if any error was raised, exit method!
				{
					return false;
				}
			}
			firstRangeNode = lastRangeNode = 0;
			//if here we are, we can process next target node
			buffer.erase(); //clean up the buffer
		}
		else if(c>='0' && c<='9')
		{
			buffer += c; //add this number into the buffer
		}
		else
		{
			*tcp() << "<ERR>: illegal characters found in request! (" << param.substr(i,1) << ")" << endl;
			return true;
		}
	}

	if(buffer.size()>0) //if the param string was not totally processed (so the buffer still contains something)...
	{
		executeSingleRequest(functionByNumber, buffer);
	}
	return true;
}

/**
 * used to execute a request (splitted in a command, in a request, in the target-nodes)
 * @param callFuntionByNumber is used to call a specific function
 * @param singleTarget is a string with the target-node
 * @return a boolean (bool) used to handle the thread status
 */
bool connHandler::executeSingleRequest(int callFuntionByNumber, string singleTarget)
{
	int  aclCode = ACL_FULL_CONTROL; //ACL_THREAD_ERROR_OCCURRED;
	long pid;
	string helpPage;
	string errorCodesPage;

	if(singleTarget=="" || singleTarget.size()==0)
	{
		singleTarget=m_rParam;
	}

	aclCode = m_globalInit->m_aclCheck->validatePermissions(m_remoteClientIP, callFuntionByNumber);
	pid = -1;

	//cout << "<ACL CHECK EXIT CODE>: " << aclCode << endl;
	if (aclCode<0)
	{
		*tcp() << "<ACL EXCEPTION> requested function cannot be executed! (code=" << aclCode << ")" << endl;
		return true;
	}

 	if(m_remoteRequest==true)
	{
		m_remoteRequest = false;
		string remoteGomdIP =  m_globalInit->m_gs->getNodeIPbyNodeID(singleTarget);
		cout << "\n[INF]: opening conn to remote GOMD (ID:" << singleTarget << ", IP:" << remoteGomdIP << ")" << endl;

		if(remoteGomdIP=="<ERROR>")
		{
			m_answerBuffer.append("<ERR> cannot get IP address for node #"+singleTarget);
			return true;
		}
		string request = "get " + m_util.intToString(callFuntionByNumber);
		cout << "\n[INF]: request will be: " << request <<endl;
		string answer = m_gomd2gomd.getInfoAndClose(remoteGomdIP,9889,request);
		m_answerBuffer =  "(NODE#:";
		m_answerBuffer += singleTarget;
		m_answerBuffer += ")";
		m_answerBuffer += answer;
		return true;
	}

	switch (callFuntionByNumber)
	{
		case 0: //option: quit
		{
			m_answerBuffer.append("closing session...");
			return false;
		}
		case 1: //option: local node id
		{

			m_answerBuffer.append("(LOCAL_ID:"+m_localNodeID+")");
			break;
		}
		case 2: //option: oM node cpus number
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("cpus",singleTarget));
			break;
		}
		case 3: //option: oM node load
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("load",singleTarget));
			break;
		}
		case 4: //option: oM node mem
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("mem",singleTarget));
			break;
		}
		case 5: //option: oM node mem
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("rmem",singleTarget));
			break;
		}
		case 6: //option: oM node speed
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("speed",singleTarget));
			break;
		}
		case 7: //option: oM node status
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("status",singleTarget));
			break;
		}
		case 8: //option: oM node mem
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("tmem",singleTarget));
			break;
		}
		case 9: //option: oM node utilization
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getInfo("util",singleTarget));
			break;
		}
		case 10: //option: oM node block status
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getLocalBlock());
			break;
		}
		case 11: //option: oM node IP
		{
			m_answerBuffer.append("(LOCAL_IP:"+m_localNodeIP+") ");
			break;
		}
		case 12: //option: oM node by PID
		{
			m_answerBuffer.append("("+singleTarget+":"+m_globalInit->m_omInfo.getNodeByPID(singleTarget)+") ");
			break;
		}
		case 13: //option: oM available nodes
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getNodesList(true));
			break;
		}
		case 14: //option: oM dead nodes
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.getDeadNodesList());
			break;
		}
		case 20: //option: oM set local block
		{
			m_answerBuffer.append(m_globalInit->m_omInfo.setLocalBlock());
			break;
		}
		case 21: //option: oM unset local block
		{
			m_answerBuffer.append("(UNSET_LOCAL_BLOCK:"+ m_globalInit->m_omInfo.setLocalNoblock()+") ");
			break;
		}
		case 22: //option: oM set local decay
		{
			m_answerBuffer.append("(SET_LOCAL_DECAY:"+ m_globalInit->m_omInfo.setLocalDecay(singleTarget)+") ");
			break;
		}
		case 30: //option: repeat last request
		{
			repeatLastRequest();
			break;
		}
		case 31: //option: repeat previous request in the history
		{
			repeatUpRequest();
			break;
		}
		case 32: //option: repeat next request in the history
		{
			repeatDownRequest();
			break;
		}
		case 33: //option: print all requests in the history
		{
			printAllRequests();
			break;
		}
		case 34: //option: get a snapshot for local node
		{
			m_answerBuffer.append(generateSnapshot());
			break;
		}
		#ifdef __LIBGTOP
		case 40: //option: get system CPUs load
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getCpusLoad());
			break;
		}
		case 41: //option: get contents of /proc/cpuinfo file
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getCpuinfoFile(""));
			break;
		}
		case 42: //option: get network load values for a specific interface (i.e. eth0,lo,...)
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getNetLoad(singleTarget.c_str()));
			break;
		}
		case 43: //option: get list of PIDs
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getPidList(NULL));
			break;
		}
		case 44: //option: get the process list
		{
			string tmp = executeCommand("get cmd ps aux");
			m_answerBuffer.append(tmp);
			//m_answerBuffer.append(m_globalInit->m_sysInfo.getProcList(NULL));
			break;
		}
		case 45: //option: get mounted fs list
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getMountList(""));
			break;
		}
		case 46: //option: get a list with usage of mounted fs
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getMountUsageList(""));
			break;
		}
		case 47: //option: get uptime and idletime
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getUptime());
			break;
		}
		case 48: //option: get contents of /proc/meminfo file
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getMeminfoFile(""));
			break;
		}
		case 49: //option: get swap usage
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getSwapUsage());
			break;
		}
		case 50: //option: get load average
		{
			m_answerBuffer.append(m_globalInit->m_sysInfo.getLoadAverage());
			break;
		}
		case 51: //option: get process state
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			m_answerBuffer.append(m_globalInit->m_sysInfo.getProcState(pid));
			break;
		}
		case 52: //option: get process state
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			m_answerBuffer.append(m_globalInit->m_sysInfo.getProcUIDAndTTY(pid));
			break;
		}
		case 53: //option: get process memory state
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			m_answerBuffer.append(m_globalInit->m_sysInfo.getProcMemory(pid));
			break;
		}
		case 54: //option: get process time state
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			m_answerBuffer.append(m_globalInit->m_sysInfo.getProcTime(pid));
			break;
		}
		case 55: //option: get process signal state
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			m_answerBuffer.append(m_globalInit->m_sysInfo.getProcSignal(pid));
			break;
		}
		#endif
		case 99: //option: get help/program usage notes
		{
			sendBannerToClient();
			helpPage = m_globalInit->m_chSupport->showHelp(m_rParam);
			m_answerBuffer.append(helpPage);
			break;
		}
		case 100: //option: start gomd autodiscovery facility
		{
			m_answerBuffer.append("(AUTODISC_ON_LAN:enabled)");
			unsigned int highestIP;
			highestIP = (singleTarget!="") ? m_util.stringToInt(singleTarget) : GOMD_SEARCHER_IP_TO_PING;
			//set the highest IP on local subnet to look for
			m_globalInit->m_gs->setMaxRangeIP(highestIP);
			m_globalInit->m_gs->startSearch();
			break;
		}
		case 101: //option: start gomd globalInitializer facility on LAN
		{
			m_answerBuffer.append("(csCOLLECTOR:enabled)");
			m_globalInit->detach();
			break;
		}
		case 102: //option: show error codes
		{
			errorCodesPage = m_globalInit->m_chSupport->showErrorCodes();
			m_answerBuffer.append(errorCodesPage);
			break;
		}
		case 103: //option: execute a command on local node
		{
			m_answerBuffer.append(executeCommand(singleTarget));
			break;
		}
		case 104: //option: execute a command in a cluster-wide context
		{
			m_answerBuffer.append(executeCWCommand(singleTarget));
			break;
		}
		case 105: //option: start gomd autodiscovery facility on nodes found by omdiscd
		{
			m_answerBuffer.append("(AUTODISC_ON_NODES_FOUND_BY_OMDISCD:enabled)");
			m_globalInit->m_gs->m_searchByOmDiscd = true;
			m_globalInit->m_gs->detach();
			break;
		}
		case 106: //option: register and dump a process
		{
			pid = m_util.stringToLong(singleTarget.c_str());
			bool r = m_globalInit->m_chpoxSupport->appendToPendingDumpsQueue(pid);
			string s = (r==true) ? "success!" : "error!";
			m_answerBuffer.append("(CHPOX_APPENDING_PROC:"+s+")");
			break;
		}
		case 107: //option: get cpu(s) temp
		{
			m_answerBuffer.append(m_globalInit->m_lmsSupport->getTemp());
			break;
		}
		case 108: //option: execute a command on local node
		{
			m_answerBuffer.append(executeCommandByNodeID(singleTarget));
			break;
		}
		default: //option: not a valid option!
		{
			m_answerBuffer.append("(<WARNING> Unknown command)");
			break;
		}
	}
	m_util.trim(m_answerBuffer); //trim the answer buffer
	return true;
}

/**
 * used to check if the current security level allows the
 * execution of a given command
 * @param command is a string with the command to be checked
 * @return an empty
 */
string connHandler::securityChecks(string command)
{
	//if strict security mode is set, check if the
	//required command is blacklisted.
	if(m_globalInit->m_security.m_isStrictSecurityEnabled==true)
	{
		if(m_globalInit->m_security.isBlackListedCommand(command))
		{
			//the command is blacklisted => return the error
			return "[ERR]: cannot execute command (blacklisted)";
		}
	}

	//verify if the command is allowed (listed in the config file)
	if(m_globalInit->m_scxCmdsbook->searchScxMapForCommand(command)==false)
	{
		return "[ERR]: cannot execute command (not listed)";
	}

	return "[NULLA OBSTAT]"; //no error occurred
}




/**
 * executes a command in a cluster-wide context
 * @param command is a string with the command to be executed
 * @return the output for the command executed (if all went right);
 *  an error message, otherwise
 */
string connHandler::executeCWCommand(string command)
{
	string ret;
	bool postPoned = false;
	string::size_type p = command.find("cwc");

	if(command.size()==0 || p==string::npos)
	{
		return "[ERR]: unspecified command.";
	}
	command = command.substr(p+4); //separing the real command from the rest

	string tmp = securityChecks(command);
	if(tmp!="[NULLA OBSTAT]")
	{
		return tmp;
	}

	command = "exec cmd " + command;
	//cout << "[SCX]: executing CW command (" << command << ")" << endl;

	map<string,string> remoteGomds = m_globalInit->m_gs->m_knownGomds;
	map<string,string>::iterator iter;
	for(iter=remoteGomds.begin(); iter!=remoteGomds.end(); ++iter)
	{
		if(string(iter->first)!=m_localNodeIP)
		{
			*tcp() << GREEN << "- sending cmd request ( " << command << ") to: " << iter->first << "..." << NORMAL << endl;
			*tcp() << GREEN << "- waiting for process output from remote gomd..." << NORMAL << endl;
			*tcp() << m_gomd2gomd.getMultilineInfoAndClose(iter->first,9889,command,"[EOA]") << endl;
		}
		else
		{
			*tcp() << GREEN << "- postponing local node..."<< NORMAL << endl;
			postPoned=true;
		}
	}

	if(postPoned==true)
	{
		*tcp() << GREEN << "- sending cmd request ( " << command << ") to: " << m_localNodeIP << " (local node)..." << NORMAL << endl;
		*tcp() << GREEN << "- waiting for process output from remote gomd..." << NORMAL << endl;
		*tcp() << m_gomd2gomd.getMultilineInfoAndClose(m_localNodeIP,9889,command,"[EOA]") << endl;
	}


	ret.append(GREEN);
	ret.append("- end of remote process output.");
	ret.append(NORMAL);
	return ret;
}


/**
 * executes a command on a specific node
 * @param command is a string with the command to be executed
 * @return the output for the command executed (if all went right);
 *  an error message, otherwise
 */
string connHandler::executeCommandByNodeID(string command)
{

	cout << "[SCX]: executeCommandByNodeID!" << endl;
	
	string ret;
	string nodeId;
	string::size_type p = command.find("con");

	if(command.size()==0 || p==string::npos)
	{
		return "[ERR]: unspecified command.";
	}
	command = command.substr(p+4); //separing the real command from the rest

	cout << command << endl;
	
	string::size_type sp = command.find("(");
	string::size_type ep = command.find(")");
	if(sp!=string::npos && ep!=string::npos)
	{
		nodeId=command.substr(sp+1,ep-1);
		command = command.substr(ep+1);
	}
	else
	{
		cout << "ERRRRRRRRRR!!!" << endl;
		return ret;
	}
	
	string tmp = securityChecks(command);
	if(tmp!="[NULLA OBSTAT]")
	{
		return tmp;
	}

	command = "exec cmd " + command;
	//cout << "[SCX]: executing CW command (" << command << ")" << endl;

	map<string,string> remoteGomds = m_globalInit->m_gs->m_knownGomds;
	map<string,string>::iterator iter;
	for(iter=remoteGomds.begin(); iter!=remoteGomds.end(); ++iter)
	{
		if(string(iter->second)==nodeId)
		{
			*tcp() << GREEN << "- sending cmd request ( " << command << ") to: " << iter->first << "..." << NORMAL << endl;
			*tcp() << GREEN << "- waiting for process output from remote gomd..." << NORMAL << endl;
			*tcp() << m_gomd2gomd.getMultilineInfoAndClose(iter->first,9889,command,"[EOA]") << endl;
		}
	}

	ret.append(GREEN);
	ret.append("- end of remote process output.");
	ret.append(NORMAL);
	return ret;
}







/**
 * executes a command on the local node only
 * @param command is a string with the command to be executed
 * @return the output for the command executed (if all went right);
 *  an error message, otherwise
 */
string connHandler::executeCommand(string command)
{
	//search fro "cmd" into the string
	string::size_type p = command.find("cmd");

	if(command.size()==0 || p==string::npos)
	{
		return "[ERR]: unspecified command.";
	}

	command = command.substr(p+4);

	string tmp = securityChecks(command);
	if(tmp!="[NULLA OBSTAT]")
	{
		return tmp;
	}

	//*tcp() << "[SCX]: executing command: " << command << endl;
	m_globalInit->m_scxCheck->tryToExecuteCommand(command, this);
	*tcp() << "[EOA]" << endl;
	return "[SCX]: command executed.";
}

/**
 * invokes the creation of a new cluster-wide snapshot and
 * send it to the remote node
 * @return a string used to communicate the end of end-of-message.
 */
string connHandler::generateSnapshot()
{
	m_clusterSnapshot.createHTMLSnapshot(m_localNodeIP, m_localNodeHostname);
	list<string>::iterator iter;
	//cout << RED << "MULTILINE SEND: starting..." << NORMAL << endl;
	//cout << RED << "MULTILINE SEND: we have to send: " << m_clusterSnapshot.m_infoList.size() << " lines." << NORMAL << endl;

	for(iter = m_clusterSnapshot.m_infoList.begin(); iter != m_clusterSnapshot.m_infoList.end(); ++iter)
	{
		if(((string)*iter).find('')!=string::npos)
		{
			continue;
		}
		//cout << GREEN << "MULTILINE SEND: " << *iter << NORMAL << endl;
		*tcp() << *iter << endl;
		this->sleep(100);
	}
	this->sleep(200);
	//cout << RED << "MULTILINE SEND: [EOI]" << NORMAL << endl;
	return "[EOI]";
}

/**
 * sends the program banner to the client
 */
void connHandler::sendBannerToClient()
{
	*tcp() << "Welcome to generic openMosix daemon at " << m_localNodeHostname << " from socket " << so << endl;
	*tcp() << "Here is node #" << m_localNodeID <<  endl;
	*tcp() << "Timeout set to: " << (m_socketTimeout/1000) << " s" <<  endl;
}

/**
 * returns the ID of the local node
 * @return a string with the ID of the local node
 */
string connHandler::getLocalIP()
{
        return m_localNodeIP;
}

/**
 * returns the last request
 * @return a string with the last request
 */
void connHandler::repeatLastRequest()
{
        m_lastRequest = m_requestHistory.getLast();
}

/**
 * returns the last "up" request
 * @return a string with the last "up" request
 */
void connHandler::repeatUpRequest()
{
        m_lastRequest = m_requestHistory.getUp();
}

/**
 * returns the last "down" request
 * @return a string with the last "down" request
 */
void connHandler::repeatDownRequest()
{
        m_lastRequest = m_requestHistory.getDown();
}

/**
 * returns the last "down" request
 * @return a string with the last "down" request
 */
void connHandler::printAllRequests()
{
	int i=0;
        cout << "[HIS] entries of request history..." << endl;

	list<string> buf(m_requestHistory.m_buffer);
	list<string>::iterator bufIter(m_requestHistory.m_bufferIter);

	for(bufIter=buf.begin() ; bufIter!=buf.end() ; ++bufIter)
	{
		++i;
		*tcp() << "#"  << i << " : " << *bufIter << endl;
	}
}
