/***************************************************************************
*                          globalInitializer.cpp  -  description
*                             -------------------
*    begin                : Sam Feb  1 17:08:52 CET 2003
*    copyright            : (C) 2003 by the gomd group at savannah.org/
*    email                : rpons@rinu.org, brian@axehind.com, jp80@libero.it, mosixview@t-online.de
****************************************************************************/

/**
 * NOTES:
 * Created by rejected
 *
 * Contains code to collect cluster-wide snapshots
 * NOTE: this class implements the SINGLETON pattern
 */

#include "globalInitializer.h"

globalInitializer* globalInitializer::m_instance = 0; //initialize pointer

/**
 * constructor for class globalInitializer.
 * @param localIP is a string with the IP adress for the local node
 * @param localHostname is a string with the hostname for the local node
 */
globalInitializer::globalInitializer(string localIP, string localHostname)
{
	m_doContinue = true;
	m_localIP = localIP;
	m_localHostname = localHostname;

	/*****************
	 * OMINFO object *
	 *****************/
	m_localNodeID = m_omInfo.getLocalID(); //set local node ID
	if(m_localNodeID==UTILS_IO_ERROR) //check if openMosix is enabled or disabled
	{
		cout << "[OMI]: openMosix not started or not configured!" << endl;
		m_omEnabled = false;
	}
	else
	{
		cout << "[OMI]: openMosix node: #" << m_localNodeID << endl;
		m_omEnabled = true;
	}

	/*********************
	 * ACL check objects *
	 *********************/
	m_aclRulebook = new aclRulebook(localHostname);
	m_aclRulebook->fillAclMap(ACL_CONF_FILE);
	m_aclRulebook->printAclMap();
	m_aclCheck = new aclCheck(m_aclRulebook->getAclMap(),m_localIP);

	/************************
	 * GOMD SEARCHER object *
	 ************************/
	m_gs = new gomdSearcher(&m_omInfo); //create the autodiscovery object (gomdSearcher)
	m_gs->setLocalIP(m_localIP); //pass it the local IP
	m_gs->setMaxRangeIP(GOMD_SEARCHER_IP_TO_PING); 	//set the highest IP on local subnet to look for
	m_gs->loadKnownGomdsMapFromFile(); 	//try reading known gomds cache file
	m_gs->appendNodeInfosToKnownGomdsMap(m_omEnabled);

	/**************
	 * SCX object *
	 **************/
	m_scxCmdsbook = new scxCmdsbook(localHostname);
	m_scxCmdsbook->fillScxMap(SCX_CONF_FILE);
	m_scxCmdsbook->printScxMap();
	m_scxCmdsbook->setScxDefaultBehaviour();
	multimap<string,string> tmp = m_scxCmdsbook->getScxMap();
	m_scxCheck = new scxCheck(tmp, localIP);

	/********************
	 * CONNHANDLER (CH) SUPPORT object *
	 ********************/
	m_chSupport = new chSupport();
	cout << "[CHS]: connHandler support stuff initialized." << endl;

	
	// THE FOLLOWING CLASSES IMPLEMENT
	
	/**************************
	 * CHPOX-SUPPORT object *
	 **************************/
	m_chpoxSupport = new chpoxSupport();
	if(m_chpoxSupport->m_isActive==true)
	{
		m_chpoxSupport->fillChpSupportConfiguration(CHPSUPPORT_CONF_FILE);
		m_chpoxSupport->printChpSupportConfiguration();
		m_chpoxSupport->detach();
	}
	
	/*********************************
	 * LOADLIMIT-SUPPORT object *
	 *********************************/
	m_llSupport = new llSupport();
	if(m_llSupport->m_isActive==true)
	{
		m_llSupport->fillLLMap(LOADLIMIT_CONF_FILE);
		m_llSupport->printLLMap();
	}
	
	/*********************************
	 * LM_SENDORS-SUPPORT object *
	 *********************************/
	m_lmsSupport = new lmsSupport();
	if(m_lmsSupport->m_isActive==true)
	{
		//OK!
	}
	
	
	
	/***************************
	 * FEATURES STATUS SUMMARY *
	 ***************************/
	cout << "[INF]: Features status map." << endl;
	cout << "       ->  openMosix : " << m_util.boolToColorString(m_omEnabled) << endl;
	cout << "       ->  loadlimit : " << m_util.boolToColorString(m_llSupport->m_isActive) << endl;
	cout << "       -> lm_sensors : " << m_util.boolToColorString(m_lmsSupport->m_isActive) << endl;
	cout << "       ->      chpox : " << m_util.boolToColorString(m_chpoxSupport->m_isActive) << endl;
	
}

/**
 * destructor for class globalInitializer
 */
globalInitializer::~globalInitializer()
{
	delete m_aclCheck;
	delete m_aclRulebook;
	delete m_chpoxSupport;
	delete m_chSupport;
	delete m_gs;
	delete m_scxCheck;
	delete m_scxCmdsbook;
}

/**
 * enable specific options for Live-CD options
 */
void globalInitializer::enableLivecdSpecificSettings()
{
	cout << BLUE << "[INF]: using specific settings for live-cd distros" << NORMAL << endl;
	m_gs->m_searchByOmDiscd = true;
	m_gs->detach();
	cout << BLUE << "[INF]: using specific settings for live-cd distros" << NORMAL << endl;

}

/*
 * read (legacy) openMosix map file, eventually update
 * the known gomd cache file and save it
 */
void globalInitializer::readOpenMosixMapFile()
{
	m_gs->parseOmMapFileAndFillKnownGomdsMap();
	m_gs->saveKnownGomdsMapToFile(); 	//save gomds cache file
}

/**
 * used to create the unique global instance of this class
 * Required for implementing the SINGLETON Design Pattern
 * @param localIP is a string with the IP adress for the local node
 * @param localHostname is a string with the hostname for the local node
 */
globalInitializer* globalInitializer::Instance(string localIP, string localHostname)
{
	if(m_instance == 0)
	{
		cout << "[INI]: instance created and ready for use." << endl;
		cout << "[INI]: initializing gomd subsystems..." << endl;
		m_instance = new globalInitializer(localIP, localHostname);
		cout << "[INI]: gomd subsystems initialized." << endl;
	}
	return m_instance;
}

/**
 * used to start the internal thread
 * (to generate the cluster-wide snapshot to file)
 */
void globalInitializer::run()
{
		this->sleep(1000);
		cout << "[INI]: initialiazing..." << endl;
		cout << "[INI]: try reading known gomds cache file..." << endl;

		/*
		 * 1) try reading the file with remote gomds cache
		 */
		m_knownGomds = m_util.parseSpecialFile(KNOWN_GOMDS_FILE,":");
		//if the file doesn't exists (and the map is empty),invoke the autodiscovery facility
		if(m_knownGomds.size()==0)
		{

			//invoke remote gomds search (detached thread)...
			cout << GREEN << "[INI]: no known gomds => starting autodiscovery..." << NORMAL << endl;

			//set the highest IP on local subnet to look for
			m_gs->setMaxRangeIP(GOMD_SEARCHER_IP_TO_PING);

			m_gs->startSearch();

			//loop until search is finished
			while(m_gs->isRunning()==true)
			{
				//give priority to another (system) thread
				this->yield();
			}

			//retry reading the file with remote gomds cache
			m_knownGomds = m_util.parseSpecialFile(KNOWN_GOMDS_FILE,":");

			//if no gomd were found...
			if(m_knownGomds.size()==0)
			{
				//exit!!!
				cout << RED << "[INI]: no known gomd found by autodiscovery!!!" << NORMAL << endl;
				return;
			}
		}

		/*
		 * 2)Print found gomds (if we are here, the known gomds map was filled)...
		 */
		cout << "[INI]: found " << m_knownGomds.size() << " remote gomd(s)"<< endl;
		map<string,string>::iterator iter;
		for(iter=m_knownGomds.begin(); iter!=m_knownGomds.end(); ++iter)
		{
			cout << "  -> " << iter->first << ":" << iter->second << endl;
		}

		/*
		 * 3)ask clusterSnapshot to each found remote gomd
		to test the saved-procs linked list. */
		string answers = "";
		for(iter=m_knownGomds.begin(); iter!=m_knownGomds.end(); ++iter)
		{
			cout << "[INI]: asking snapshot to gomd: "<< iter->first <<  endl;
			answers += m_g2g.getMultilineInfoAndClose(iter->first,9889,"get 34","[EOI]");
		}

		/*
		 * 4)make the cluster-wide HTML report
		 */
		string htmlReport = "";
		htmlReport += "<HTML>\n";
		htmlReport += "<HEAD>\n";
		htmlReport += "<TITLE>clusterSnapshot HTML report</TITLE>\n";
		htmlReport += "</HEAD>\n";
		htmlReport += "<BODY>\n";
		htmlReport += answers;
		htmlReport += "</TABLE>\n";
		htmlReport += "</FONT>\n";
		htmlReport += "</BODY>\n";
		htmlReport += "</HTML>\n";

		/*
		 * 5)save cluster-wide HTML report to file
		 */
		cout << "[INI]: writing HTML report file ("<< CS_HTML_REPORT_FILE <<  ")." << endl;
		m_util.writeToFile(CS_HTML_REPORT_FILE,htmlReport);

		/*
		 * 6)exiting... ;)
		 */
		Thread::exit();
}

/**
 * try registering (three times) the INIT process via the chpox mechanism
 * first time an entry will be created; the following two times that entry
 * will be updated.
 */
void globalInitializer::testRegisterProcess()
{
	unsigned int dumpFileID = 1;
	pid_t pidToBeDumped=1;
	string dumpFilePath;
	
	
	list<string> cmdsList = m_util.parseCommand("xxx");
	list<string> ret = m_util.execCommandAndReturnChildOutput(cmdsList, NULL);  //execute the command (second param is set to NULL => no real-time output)
	m_util.printLinkedList(ret,"","");

	cout << RED << "\n" << endl;
	cout << "***************************************" << endl;
	cout << "*    TESTING CHPOX SUPPORT (ALPHA)    *" << endl;
	cout << "***************************************\n" << endl;
	cout <<  "[TST]: this code is experimental! :)" << endl;
	cout << BLUE << "[TST]: this test is simple:" << endl;
	cout << "[TST]: 1) you'll be asked for the PID of a process." << endl;
	cout << "[TST]: 2) the given process will be registered " << CHPOX_MAX_DUMPS_PER_PROCESS << " times (with all the required libs)" << endl;
	cout << "[TST]: 3) you'll see the list of dump files." << endl;
	cout << "[TST]: 4) you'll be asked to choose the dump file for restoring" << endl;
	cout << "[TST]: 5) the given will be killed (SIGKILL) then will be restored by the choosen dump file.\n" << NORMAL << endl;

	cout << RED "[TST]: please enter the PID of process to be checkpointed: " << NORMAL;
	cin >> pidToBeDumped;

	for(int i=0;i<CHPOX_MAX_DUMPS_PER_PROCESS;i++)
	{
		cout << RED << "\n[TST]: round "<< (i+1) << "/"<< CHPOX_MAX_DUMPS_PER_PROCESS << NORMAL << endl;
		cout << "[TST]: appending process with pid=" << pidToBeDumped << endl;
		m_chpoxSupport->appendToPendingDumpsQueue(pidToBeDumped);
	}

	for(int i=30;i>0;i--)
	{
		printf("[TST]: countdown before checking the results: %d \r",i);
		fflush(stdout);
		sleep(1000);
	}
	cout << RED << "\n[TST]: list of dump files for the registered process (PID=" << pidToBeDumped << "):" <<NORMAL << endl;
	m_chpoxSupport->printRegisteredProcList();

	cout << RED "[TST]: please enter the id of the dump file to be used for restoring the process:" << NORMAL;
	cin >> dumpFileID;
	dumpFilePath =  m_chpoxSupport->restoreProcess(pidToBeDumped, dumpFileID, true);
}

